#include <dune/solvers/common/arithmetic.hh>

template <class Vector, class Matrix, class Function, size_t dim>
Newmark<Vector, Matrix, Function, dim>::Newmark(
    Matrix const &_A, Matrix const &_M, Matrix const &_C,
    Vector const &_u_initial, Vector const &_v_initial,
    Vector const &_a_initial, Dune::BitSetVector<dim> const &_dirichletNodes,
    Function const &_dirichletFunction)
    : A(_A),
      M(_M),
      C(_C),
      u(_u_initial),
      v(_v_initial),
      a(_a_initial),
      dirichletNodes(_dirichletNodes),
      dirichletFunction(_dirichletFunction) {}

template <class Vector, class Matrix, class Function, size_t dim>
void Newmark<Vector, Matrix, Function, dim>::nextTimeStep() {
  a_o = a;
  v_o = v;
  u_o = u;
}

template <class Vector, class Matrix, class Function, size_t dim>
void Newmark<Vector, Matrix, Function, dim>::setup(Vector const &ell,
                                                   double _tau,
                                                   double relativeTime,
                                                   Vector &rhs, Vector &iterate,
                                                   Matrix &AM) {
  postProcessCalled = false;
  dirichletFunction.evaluate(relativeTime, dirichletValue);
  tau = _tau;

  /* We start out with the formulation

       M a + C v + A u = ell

     Newmark means

       a1 = 2/tau ( v1 - v0 ) - a0
       u1 = tau/2 ( v1 + v0 ) + u0

     in summary, we get at time t=1

       M [2/tau ( u1 - u0 ) - a0] + C v1
       + A [tau/2 ( v1 + v0 ) + u0] = ell

     or

       2/tau M v1 + C v1 + tau/2 A v1
       = [2/tau M + C + tau/2 A] v1
       = ell + 2/tau M v0 + M a0
       - tau/2 A v0 - A u0
  */

  // set up LHS (for fixed tau, we'd only really have to do this once)
  {
    Dune::MatrixIndexSet indices(A.N(), A.M());
    indices.import(A);
    indices.import(M);
    indices.import(C);
    indices.exportIdx(AM);
  }
  AM = 0.0;
  Arithmetic::addProduct(AM, 2.0 / tau, M);
  Arithmetic::addProduct(AM, 1.0, C);
  Arithmetic::addProduct(AM, tau / 2.0, A);

  // set up RHS
  {
    rhs = ell;
    Arithmetic::addProduct(rhs, 2.0 / tau, M, v_o);
    Arithmetic::addProduct(rhs, M, a_o);
    Arithmetic::subtractProduct(rhs, tau / 2.0, A, v_o);
    Arithmetic::subtractProduct(rhs, A, u_o);
  }

  iterate = v_o;

  for (size_t i = 0; i < dirichletNodes.size(); ++i)
    for (size_t j = 0; j < dim; ++j)
      if (dirichletNodes[i][j])
        iterate[i][j] = (j == 0) ? dirichletValue : 0;
}

template <class Vector, class Matrix, class Function, size_t dim>
void Newmark<Vector, Matrix, Function, dim>::postProcess(
    Vector const &iterate) {
  postProcessCalled = true;

  v = iterate;

  // u1 = tau/2 ( v1 + v0 ) + u0
  u = u_o;
  Arithmetic::addProduct(u, tau / 2.0, v);
  Arithmetic::addProduct(u, tau / 2.0, v_o);

  // a1 = 2/tau ( v1 - v0 ) - a0
  a = 0;
  Arithmetic::addProduct(a, 2.0 / tau, v);
  Arithmetic::subtractProduct(a, 2.0 / tau, v_o);
  Arithmetic::subtractProduct(a, 1.0, a_o);
}

template <class Vector, class Matrix, class Function, size_t dim>
void Newmark<Vector, Matrix, Function, dim>::extractDisplacement(
    Vector &displacement) const {
  if (!postProcessCalled)
    DUNE_THROW(Dune::Exception, "It seems you forgot to call postProcess!");

  displacement = u;
}

template <class Vector, class Matrix, class Function, size_t dim>
void Newmark<Vector, Matrix, Function, dim>::extractVelocity(Vector &velocity)
    const {
  if (!postProcessCalled)
    DUNE_THROW(Dune::Exception, "It seems you forgot to call postProcess!");

  velocity = v;
}

template <class Vector, class Matrix, class Function, size_t dim>
void Newmark<Vector, Matrix, Function, dim>::extractOldVelocity(
    Vector &velocity) const {
  velocity = v_o;
}

template <class Vector, class Matrix, class Function, size_t dim>
std::shared_ptr<TimeSteppingScheme<Vector, Matrix, Function, dim>>
Newmark<Vector, Matrix, Function, dim>::clone() const {
  return std::make_shared<Newmark<Vector, Matrix, Function, dim>>(*this);
}
